"""Module defining the low level UOSImplementation for serial port devices."""
import platform
from logging import getLogger as Log
from time import sleep, time_ns

import serial
from serial.serialutil import SerialException
from serial.tools import list_ports

from uoshardware.abstractions import ComResult, UOSInterface

if platform.system() == "Linux":
    import termios  # pylint: disable=E0401
else:
    pass


class Serial(UOSInterface):
    """Pyserial class that handles reading / writing to ports.

    :ivar _device: Holds the pyserial device once opened. None if not opened.
    :ivar _connection: Holds the standard connection string 'Interface'|'OS Connection String.
    :ivar _port: Holds the port class, none type if device not instantiated.
    :ivar _kwargs: Additional keyword arguments as defined in the documentation.
    """

    _device = None

    _connection = ""
    _port = None
    _kwargs = {}

    def __init__(self, connection: str, **kwargs):
        """Create an NPCSerialPort device.

        :param connection: OS connection string for the serial port.
        """
        self._connection = connection
        self._port = self.check_port_exists(connection)
        self._kwargs = kwargs
        if self._port is None:
            Log(__name__).error("%s port does not exist", connection)
        else:
            Log(__name__).debug("%s located", self._port)

    def open(self):
        """Open a connection to the port and creates the device object.

        :return: Success boolean.
        """
        try:
            self._port = self.check_port_exists(self._connection)
            if self._port is None:
                Log(__name__).error(
                    "%s device was not present to open", self._connection
                )
                return False
            self._device = serial.Serial()
            self._device.port = self._connection
            if "baudrate" in self._kwargs:
                self._device.baudrate = self._kwargs["baudrate"]
            if platform.system() == "Linux":  # DTR transient workaround for Unix
                Log(__name__).debug("Linux platform found so using DTR workaround")
                with open(self._connection, encoding="b") as port:
                    attrs = termios.tcgetattr(port)
                    attrs[2] = attrs[2] & ~termios.HUPCL
                    termios.tcsetattr(port, termios.TCSAFLUSH, attrs)
            else:  # DTR transient workaround for Windows
                self._device.dtr = False
            self._device.open()
            Log(__name__).debug("%s opened successfully", self._port.device)
            return True
        except (SerialException, FileNotFoundError) as exception:
            Log(__name__).error(
                "Opening %s threw error %s",
                self._port.device if self._port is not None else "None",
                str(exception),
            )
            if (
                exception.errno == 13
            ):  # permission denied another connection open to this device.
                Log(__name__).error(
                    "Cannot open connection, account has insufficient permissions."
                )
            self._device = None
            return False

    def close(self):
        """Close the serial connection and clear the device.

        :return: Success boolean.
        """
        if not self.check_open():
            return True  # already closed
        try:
            self._device.close()
        except SerialException as exception:
            Log(__name__).debug("Closing the connection threw error %s", str(exception))
            self._device = None
            return False
        Log(__name__).debug("Connection closed successfully")
        self._device = None
        return True

    def execute_instruction(self, address, payload):
        """Build and execute a new instruction packet.

        :param address: An 8-bit unsigned integer of the UOS subsystem targeted by the instruction.
        :param payload: A tuple containing the uint8 parameters of the UOS instruction.
        :return: Tuple containing a status boolean and index 0 and a result-set dict at index 1.
        """
        if not self.check_open():
            return ComResult(False, exception="Connection must be opened first.")
        packet = self.get_npc_packet(to_addr=address, from_addr=0, payload=payload)
        Log(__name__).debug("packet formed %s", packet)
        try:  # Send the packet.
            num_bytes = self._device.write(packet)
            self._device.flush()
            Log(__name__).debug("Sent %s bytes of data", num_bytes)
        except serial.SerialException as exception:
            return ComResult(False, exception=str(exception))
        finally:
            self._device.reset_output_buffer()
        return ComResult(num_bytes == len(packet))

    def read_response(self, expect_packets: int, timeout_s: float):
        """Read ACK and response packets from the serial device.

        :param expect_packets: How many packets including ACK to expect.
        :param timeout_s: The maximum time this function will wait for data.
        :return: ComResult object.
        """
        response_object = ComResult(False)
        if not self.check_open():
            return response_object
        start_ns = time_ns()
        packet = []
        byte_index = -1  # tracks the byte position index of the current packet
        packet_index = 0  # tracks the packet number being received 0 = ACK
        try:
            while (
                timeout_s * 1000000000
            ) > time_ns() - start_ns and byte_index > -2:  # read until packet or timeout
                num_bytes = self._device.in_waiting
                for _ in range(num_bytes):
                    byte_in = self._device.read(1)
                    byte_index, packet = self.decode_and_capture(
                        byte_index, byte_in, packet
                    )
                    if byte_index == -2:
                        if packet_index == 0:
                            response_object.ack_packet = packet
                        else:
                            response_object.rx_packets.append(packet)
                        packet_index += 1
                        if expect_packets == packet_index:
                            break
                        byte_index = -1
                        packet = []
                    byte_index += 1
                sleep(0.05)  # Don't churn CPU cycles waiting for data
            Log(__name__).debug("Packet received %s", packet)
            if expect_packets != packet_index or len(packet) < 6 or byte_index != -2:
                response_object.rx_packets.append(packet)
                response_object.exception = "did not receive all the expected data"
                return response_object
            response_object.status = True
            return response_object
        except serial.SerialException as exception:
            response_object.exception = str(exception)
            return response_object

    def hard_reset(self):
        """Manually drive the DTR line low to reset the device.

        :return: Tuple containing a status boolean and index 0 and a result-set dict at index 1.
        """
        if not self.check_open():
            return ComResult(False, exception="Connection must be open first.")
        Log(__name__).debug("Resetting the device using the DTR line")
        self._device.dtr = not self._device.dtr
        sleep(0.2)
        self._device.dtr = not self._device.dtr
        return ComResult(True)

    def check_open(self) -> bool:
        """Test if the connection is open.

        :return: Boolean, true if open.
        """
        if self._device is None:
            return False
        return True

    def __repr__(self):
        """Representation of object.

        :return: String containing connection, port and device.
        """
        return (
            f"<NPCSerialPort(_connection='{self._connection}', _port={self._port}, "
            f"_device={self._device})>"
        )

    @staticmethod
    def decode_and_capture(
        byte_index: int, byte_in: bytes, packet: list
    ) -> (int, list):
        """Parser takes in a byte and vets it against UOS packet.

        :param byte_index: The index of the last 'valid' byte found.
        :param byte_in: The current byte for inspection.
        :param packet: The current packet of validated bytes.
        :return: Tuple containing the updated byte index and updated packet.
        """
        if byte_index == -1:  # start symbol
            if byte_in == b">":
                byte_index += 1
        if byte_index >= 0:
            Log(__name__).debug(
                "read %s byte index = %s",
                byte_in,
                byte_index,
            )
            payload_len = packet[3] if len(packet) > 3 else 0
            if byte_index == 3 + 2 + payload_len:  # End packet symbol
                if byte_in == b"<":
                    byte_index = -2  # packet complete
                    Log(__name__).debug("Found end packet symbol")
                else:  # Errored data
                    byte_index = -1
                    packet = []
            packet.append(int.from_bytes(byte_in, byteorder="little"))
        return byte_index, packet

    @staticmethod
    def enumerate_devices():
        """Get the available ports on the system."""
        return [Serial(port.device) for port in list_ports.comports()]

    @staticmethod
    def check_port_exists(device: str):
        """Check if serial device is available on system.

        :param device: OS connection string for the serial port.
        :return: The port device class if it exists, else None.
        """
        ports = list_ports.comports()
        for port in ports:
            if device in port.device:
                return port
        return None



"""Package is used as a simulated UOSInterface for test purposes."""
from typing import Tuple

from uoshardware.abstractions import UOS_SCHEMA, ComResult, UOSInterface


class Stub(UOSInterface):
    """Class can be used as a low level test endpoint."""

    def __init__(self, connection: str, errored: int = 0):
        """Instantiate an instance of the test stub."""
        self.__packet_buffer = []
        self.__open = False
        self.errored = errored
        self.connection = connection

    def execute_instruction(
        self,
        address: int,
        # Dead code false positive as this is over-riding an interface.
        payload: Tuple[int, ...],  # dead: disable
    ) -> ComResult:
        """Simulate executing an instruction on a UOS endpoint.

        Should check whether the last instruction was valid and store
        it. This will allow read response to provide more realistic
        responses.
        """
        if not self.__open:
            return ComResult(status=False, exception="Port must be opened first.")
        for _, function in UOS_SCHEMA.items():
            for vol in function.address_lut:
                if function.address_lut[vol] == address:
                    if function.ack:
                        self.__packet_buffer.append(
                            self.get_npc_packet(0, address, tuple([0]))
                        )
                    for rx_packet in function.rx_packets_expected:
                        self.__packet_buffer.append(
                            self.get_npc_packet(
                                0, address, tuple(0 for _ in range(rx_packet))
                            )
                        )
                    return ComResult(True)
        return ComResult(False)

    # Dead code detection false positive due to abstract interface.
    def read_response(
        self, expect_packets: int, timeout_s: float  # dead: disable
    ) -> ComResult:
        """Simulate gathering the response from an instruction.

        Should have already executed an instruction. If no response is
        generated by instruction will error accordingly.
        """
        if not self.__open:
            return ComResult(status=False, exception="Port must be opened first.")
        result = ComResult(False)
        if len(self.__packet_buffer) > 0:
            result.ack_packet = self.__packet_buffer.pop(0)
            result.status = True
        for _ in self.__packet_buffer:
            result.rx_packets.append(self.__packet_buffer.pop(0))
        return result

    def hard_reset(self) -> ComResult:
        """Override base prototype, simulates reset."""
        if not self.__open:
            return ComResult(status=False, exception="Port must be opened first.")
        return ComResult(status=True)

    def open(self) -> bool:
        """Override base prototype, simulates opening a connection."""
        if len(self.connection) > 0:
            self.__open = True
            return True
        return False

    def close(self) -> bool:
        """Override base prototype, simulates close a connection."""
        self.__open = False
        return self.errored == 0

    @staticmethod
    def enumerate_devices() -> []:
        """Return a list of test stubs implemented in the interface."""
        return [Stub("STUB")]  # The test stub is always available



"""The high level interface for communicating with UOS devices."""
from enum import Enum
from logging import FileHandler, Formatter, getLogger
from pathlib import Path

__author__ = "Steve Richardson (Creating Null)"
__copywright__ = f"2022, {__author__}"
# Semantic Versioning, MAJOR.MINOR.PATCH[-'pre-release-type'.'num']
__version__ = "0.1.0"
# Dead code false positive as this constant is for use outside primary project.
PROJECT = "UOS Hardware"  # dead: disable


# Dead code false positive as this enum if for client usage.
class Persistence(Enum):
    """Volatility levels that can be used in UOS instructions."""

    NONE = 0
    RAM = 1
    EEPROM = 2  # dead: disable


# Dead code false positive as interface intended to be used by client.
def configure_logs(name: str, level: int, base_path: Path):  # dead: disable
    """Per-package logs must be manually configured to prefix correctly."""
    logger = getLogger(name)
    logger.setLevel(level)
    # Don't capture to console as custom messages only, root logger captures stderr
    logger.propagate = False
    log_dir = Path(base_path.joinpath(Path("logs/")))
    if not log_dir.exists():
        log_dir.mkdir()
    file_handler = FileHandler(log_dir.joinpath(Path(name + ".log")))
    file_handler.setFormatter(
        Formatter("%(asctime)s : %(levelname)s : %(name)s : %(message)s")
    )
    logger.addHandler(file_handler)


class UOSError(Exception):
    """Base class exception for all UOS Interface Errors."""


class UOSUnsupportedError(UOSError):
    """Exception for attempting an unknown / unsupported action."""


class UOSCommunicationError(UOSError):
    """Exception while communicating with a UOS Device."""



"""Module defining the base class and static func for interfaces."""
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from functools import lru_cache
from typing import Dict, List, Tuple

from uoshardware import Persistence, UOSUnsupportedError


@dataclass
class UOSFunction:
    """Defines auxiliary information for UOS commands in the schema."""

    address_lut: dict
    ack: bool
    rx_packets_expected: list = field(default_factory=list)
    pin_requirements: list = None


UOS_SCHEMA = {
    "set_gpio_output": UOSFunction(
        address_lut={0: 64},
        ack=True,
        pin_requirements=["gpio_out"],
    ),
    "get_gpio_input": UOSFunction(
        address_lut={0: 64},
        ack=True,
        rx_packets_expected=[1],
        pin_requirements=["gpio_in"],
    ),
    "get_adc_input": UOSFunction(
        address_lut={0: 85},
        ack=True,
        rx_packets_expected=[2],
        pin_requirements=["adc_in"],
    ),
    "reset_all_io": UOSFunction(address_lut={0: 68}, ack=True),
    "hard_reset": UOSFunction(address_lut={0: -1}, ack=False),
    "get_system_info": UOSFunction(
        address_lut={0: 250}, ack=True, rx_packets_expected=[6]
    ),
    "get_gpio_config": UOSFunction(
        address_lut={0: 251},
        ack=True,
        rx_packets_expected=[2],
        pin_requirements=[],
    ),
}


@dataclass
class ComResult:
    """Containing the data structure used to capture UOS results."""

    status: bool
    exception: str = ""
    ack_packet: List = field(default_factory=list)
    rx_packets: List = field(default_factory=list)
    aux_data: Dict = field(default_factory=dict)


@dataclass
class InstructionArguments:
    """Containing the data structure used to generalise UOS arguments."""

    device_function_lut: Dict = field(default_factory=dict)
    payload: tuple = ()
    expected_rx_packets: int = 1
    check_pin: int = None
    volatility: Persistence = Persistence.RAM


class UOSInterface(metaclass=ABCMeta):
    """Base class for low level UOS interfaces classes to inherit."""

    # Dead code suppression used as abstract interfaces are false positives.
    @abstractmethod
    def execute_instruction(
        self, address: int, payload: Tuple[int, ...]  # dead: disable
    ) -> ComResult:
        """Abstract method for executing instructions on UOSInterfaces.

        :param address: An 8-bit unsigned integer of the UOS subsystem targeted by the instruction.
        :param payload: A tuple containing the uint8 parameters of the UOS instruction.
        :returns: ComResult object.
        :raises: UOSUnsupportedError if the interface hasn't been built correctly.
        """
        raise UOSUnsupportedError(
            f"UOSInterfaces must over-ride {UOSInterface.execute_instruction.__name__} prototype."
        )

    @abstractmethod
    def read_response(
        self, expect_packets: int, timeout_s: float  # dead: disable
    ) -> ComResult:
        """Read ACK and Data packets from a UOSInterface.

        :param expect_packets: How many packets including ACK to expect
        :param timeout_s: The maximum time this function will wait for data.
        :return: COM Result object.
        :raises: UOSUnsupportedError if the interface hasn't been built correctly.
        """
        raise UOSUnsupportedError(
            f"UOSInterfaces must over-ride {UOSInterface.read_response.__name__} prototype."
        )

    @abstractmethod
    def hard_reset(self) -> ComResult:
        """UOS loop reset functionality should be as hard a reset as possible.

        :return: COM Result object.
        """
        raise UOSUnsupportedError(
            f"UOSInterfaces must over-ride {UOSInterface.hard_reset.__name__} prototype"
        )

    @abstractmethod
    def open(self) -> bool:
        """Abstract method for opening a connection to a UOSInterface.

        :return: Success boolean.
        :raises: UOSUnsupportedError if the interface hasn't been built correctly.
        """
        raise UOSUnsupportedError(
            f"UOSInterfaces must over-ride {UOSInterface.open.__name__} prototype."
        )

    @abstractmethod
    def close(self) -> bool:
        """Abstract method for closing a connection to a UOSInterface.

        :return: Success boolean.
        :raises: UOSUnsupportedError if the interface hasn't been built correctly.
        """
        raise UOSUnsupportedError(
            f"UOSInterfaces must over-ride {UOSInterface.close.__name__} prototype."
        )

    @staticmethod
    @abstractmethod
    def enumerate_devices() -> []:
        """Return a list of UOSDevices visible to the driver.

        :return: A list of possible UOSInterfaces on the server.
        :raises: UOSUnsupportedError if the interface hasn't been built correctly.
        """
        raise UOSUnsupportedError(
            f"UOSInterfaces must over-ride {UOSInterface.enumerate_devices.__name__} prototype."
        )

    @staticmethod
    @lru_cache(maxsize=100)
    def get_npc_packet(to_addr: int, from_addr: int, payload: Tuple[int, ...]) -> bytes:
        """Generate a standardised NPC binary packet.

        :param to_addr: An 8-bit unsigned integer of the UOS subsystem targeted by the instruction.
        :param from_addr: An 8-bit unsigned integer of the host system, usually 0.
        :param payload: A tuple containing the unsigned 8-bit integers of the command.
        :return: NPC packet as a bytes object. No bytes returned on fault.
        """
        if (
            to_addr < 256 and from_addr < 256 and len(payload) < 256
        ):  # check input is possible to parse
            packet_data = [to_addr, from_addr, len(payload)] + list(payload)
            lrc = UOSInterface.get_npc_checksum(packet_data)
            return bytes(
                [0x3E, packet_data[0], packet_data[1], len(payload)]
                + list(payload)
                + [lrc, 0x3C]
            )
        return bytes([])

    @staticmethod
    def get_npc_checksum(packet_data: list[int]) -> int:
        """Generate a NPC LRC checksum.

        :param packet_data: List of the uint8 values from an NPC packet.
        :return: NPC checksum as a 8-bit integer.
        """
        lrc = 0
        for byte in packet_data:
            lrc = (lrc + byte) & 0xFF
        return ((lrc ^ 0xFF) + 1) & 0xFF


@dataclass
class Pin:
    """Defines supported features of the pin."""

    # pylint: disable=too-many-instance-attributes
    # Due to the nature of embedded pin complexity.

    gpio_out: bool = False
    gpio_in: bool = False
    dac_out: bool = False
    pwm_out: bool = False
    adc_in: bool = False
    pull_up: bool = False
    pull_down: bool = False
    pc_int: bool = False
    hw_int: bool = False
    timer: dict = field(default_factory=dict)
    comp: dict = field(default_factory=dict)
    spi: dict = field(default_factory=dict)
    i2c: dict = field(default_factory=dict)


@dataclass
class Device:
    """Define an implemented UOS device dictionary."""

    name: str
    interfaces: list
    functions_enabled: dict
    digital_pins: dict = field(default_factory=dict)
    analogue_pins: dict = field(default_factory=dict)
    aux_params: dict = field(default_factory=dict)

    def get_compatible_pins(self, function_name: str) -> {}:
        """Return a dict of pin objects that are suitable for a function.

        :param function_name: the string name of the UOS Schema function.
        :return: Dict of pin objects, keyed on pin index.
        """
        if function_name not in UOS_SCHEMA:
            raise UOSUnsupportedError(f"UOS function {function_name} doesn't exist.")
        requirements = UOS_SCHEMA[function_name].pin_requirements
        if requirements is None:  # pins are not relevant to this function
            return {}
        pin_dict = self.analogue_pins if "adc_in" in requirements else self.digital_pins
        return {
            pin_name: pin
            for pin_name, pin in pin_dict.items()
            if all(hasattr(pin, requirement) for requirement in requirements)
        }



"""Provides the HAL layer for communicating with the hardware."""
from logging import getLogger as Log
from typing import Union

from uoshardware import Persistence, UOSCommunicationError, UOSUnsupportedError
from uoshardware.abstractions import ComResult, Device, InstructionArguments
from uoshardware.devices import get_device_definition
from uoshardware.interface import Interface
from uoshardware.interface.serial import Serial
from uoshardware.interface.stub import Stub


class UOSDevice:
    """Class for high level object-orientated control of UOS devices.

    :ivar identity: The type of device, this is must have a valid device in the config.
    :ivar connection: Compliant connection string for identifying the device and interface.
    :ivar device: Device definitions as parsed from a compatible ini.
    :ivar __kwargs: Connection specific / optional parameters.
    :ivar __device_interface: Lower level communication protocol layer.
    """

    identity = ""
    address = ""
    device = Device
    __kwargs = {}
    __device_interface = None

    def __init__(
        self,
        identity: Union[str, Device],
        address: str,
        interface: Interface = Interface.SERIAL,
        **kwargs,
    ):
        """Instantiate a UOS device instance for communication.

        :param identity: Specify the type of device, this must exist in the device dictionary.
        :param address: Compliant connection string for identifying the device and interface.
        :param interface: Set the type of interface to use for communication.
        :param kwargs: Additional optional connection parameters as defined in documentation.
        """
        self.identity = identity
        self.address = address
        if isinstance(identity, str):
            self.device = get_device_definition(identity)
        else:
            self.device = identity
        self.__kwargs = kwargs
        if self.device is None:
            raise UOSUnsupportedError(
                f"'{self.identity}' does not have a valid look up table"
            )
        if interface == Interface.SERIAL and Interface.SERIAL in self.device.interfaces:
            self.__device_interface = Serial(
                address,
                baudrate=self.device.aux_params["default_baudrate"],
            )
        elif interface == Interface.STUB and Interface.STUB in self.device.interfaces:
            self.__device_interface = Stub(
                connection=address,
                errored=(kwargs["errored"] if "errored" in kwargs else False),
            )
        else:
            raise UOSCommunicationError(
                f"Could not correctly open a connection to {self.identity} - {self.address}"
            )
        if not self.is_lazy():  # eager connections open when they are created
            self.open()
        Log(__name__).debug("Created device %s", self.__device_interface.__repr__())

    def set_gpio_output(
        self, pin: int, level: int, volatility: Persistence = Persistence.NONE
    ) -> ComResult:
        """Set a pin to digital output mode and sets a level on that pin.

        :param pin: The numeric number of the pin as defined in the dictionary for that device.
        :param level: The output level, 0 - low, 1 - High.
        :param volatility: How volatile should the command be, use constants from uoshardware.
        :return: ComResult object.
        """
        return self.__execute_instruction(
            UOSDevice.set_gpio_output.__name__,
            InstructionArguments(
                device_function_lut=self.device.functions_enabled,
                payload=(pin, 0, level),
                check_pin=pin,
                volatility=volatility,
            ),
        )

    def get_gpio_input(
        self, pin: int, level: int, volatility: Persistence = Persistence.NONE
    ) -> ComResult:
        """Read a GPIO pins level from device and returns the value.

        :param pin: The numeric number of the pin as defined in the dictionary for that device.
        :param level: Not used currently, future will define pull-up state.
        :param volatility: How volatile should the command be, use constants from uoshardware.
        :return: ComResult object.
        """
        return self.__execute_instruction(
            UOSDevice.get_gpio_input.__name__,
            InstructionArguments(
                device_function_lut=self.device.functions_enabled,
                payload=(pin, 1, level),
                expected_rx_packets=2,
                check_pin=pin,
                volatility=volatility,
            ),
        )

    def get_adc_input(
        self,
        pin: int,
        volatility: Persistence = Persistence.NONE,
    ) -> ComResult:
        """Read the current 10 bit ADC value.

        :param pin: The index of the analogue pin to read
        :param volatility: How volatile should the command be, use constants from uoshardware.
        :return: ComResult object containing the ADC readings.
        """
        return self.__execute_instruction(
            UOSDevice.get_adc_input.__name__,
            InstructionArguments(
                device_function_lut=self.device.functions_enabled,
                payload=tuple([pin]),
                expected_rx_packets=2,
                check_pin=pin,
                volatility=volatility,
            ),
        )

    def get_system_info(self) -> ComResult:
        """Read the UOS version and device type.

        :return: ComResult object containing the system information.
        """
        return self.__execute_instruction(
            UOSDevice.get_system_info.__name__,
            InstructionArguments(
                device_function_lut=self.device.functions_enabled,
                expected_rx_packets=2,
            ),
        )

    def get_gpio_config(self, pin: int) -> ComResult:
        """Read the configuration for a digital pin on the device.

        :param pin: Defines the pin for config querying.
        :return: ComResult object containing the system information.
        """
        return self.__execute_instruction(
            UOSDevice.get_gpio_config.__name__,
            InstructionArguments(
                device_function_lut=self.device.functions_enabled,
                payload=tuple([pin]),
                expected_rx_packets=2,
                check_pin=pin,
            ),
        )

    def reset_all_io(self) -> ComResult:
        """Execute the reset IO at the defined volatility level."""
        return self.__execute_instruction(
            UOSDevice.reset_all_io.__name__,
            InstructionArguments(device_function_lut=self.device.functions_enabled),
        )

    def hard_reset(self) -> ComResult:
        """Hard reset functionality for the UOS Device."""
        return self.__execute_instruction(
            UOSDevice.hard_reset.__name__,
            InstructionArguments(device_function_lut=self.device.functions_enabled),
        )

    def open(self):
        """Connect to the device, explict calls are normally not required.

        :raises: UOSCommunicationError - Problem opening a connection.
        """
        if not self.__device_interface.open():
            raise UOSCommunicationError(
                "There was an error opening a connection to the device."
            )

    def close(self):
        """Release connection, must be called explicitly if loading is eager.

        :raises: UOSCommunicationError - Problem closing the connection to an active device.
        """
        if not self.__device_interface.close():
            raise UOSCommunicationError(
                "There was an error closing a connection to the device"
            )

    def __execute_instruction(
        self,
        function_name: str,
        instruction_data: InstructionArguments,
        retry: bool = True,
    ) -> ComResult:
        """Execute a generic UOS function and get the result.

        :param function_name: The name of the function in the OOL.
        :param instruction_data: device_functions from the LUT, payload ect.
        :param retry: Allows the instruction to retry execution when fails.
        :return: ComResult object
        :raises: UOSUnsupportedError if function is not possible on the loaded device.
        """
        if function_name not in self.device.functions_enabled or (
            instruction_data.check_pin is not None
            and instruction_data.check_pin
            not in self.device.get_compatible_pins(function_name)
        ):
            Log(__name__).debug(
                "Known functions %s", str(self.device.functions_enabled.keys())
            )
            raise UOSUnsupportedError(
                f"{function_name}({instruction_data.volatility.name}) "
                f"has not been implemented for {self.identity}"
            )
        rx_response = ComResult(False)
        if self.is_lazy():  # Lazy loaded
            self.open()
        if (
            instruction_data.device_function_lut[function_name][
                instruction_data.volatility.value
            ]
            >= 0
        ):  # a normal instruction
            tx_response = self.__device_interface.execute_instruction(
                instruction_data.device_function_lut[function_name][
                    instruction_data.volatility.value
                ],
                instruction_data.payload,
            )
            if tx_response.status:
                rx_response = self.__device_interface.read_response(
                    instruction_data.expected_rx_packets, 2
                )
                if rx_response.status:
                    # validate checksums on all packets
                    for count in range(len(rx_response.rx_packets) + 1):
                        current_packet = (
                            rx_response.ack_packet
                            if count == 0
                            else rx_response.rx_packets[count - 1]
                        )
                        computed_checksum = self.__device_interface.get_npc_checksum(
                            current_packet[1:-2]
                        )
                        Log(__name__).debug(
                            "Calculated checksum %s must match rx %s",
                            computed_checksum,
                            current_packet[-2],
                        )
                        rx_response.status = rx_response.status & (
                            computed_checksum == current_packet[-2]
                        )
        else:  # run a special action
            rx_response = getattr(self.__device_interface, function_name)()
        if self.is_lazy():  # Lazy loaded
            self.close()
        if (
            not rx_response.status and retry
        ):  # allow one retry per instruction due to DTR resets
            return self.__execute_instruction(function_name, instruction_data, False)
        return rx_response

    def is_lazy(self) -> bool:
        """Check the loading type of the device lazy or eager.

        :return: Boolean, true is lazy.
        """
        if "loading" not in self.__kwargs or self.__kwargs["loading"].upper() == "LAZY":
            return True
        return False

    def __repr__(self):
        """Representation of the UOS device.

        :return: String containing connection and identity of the device
        """
        return (
            f"<UOSDevice(address='{self.address}', identity='{self.identity}', "
            f"device={self.device}, __device_interface='{self.__device_interface}', "
            f"__kwargs={self.__kwargs})>"
        )



"""Package defining io interfaces used for NPC Comms."""
from enum import Enum

from uoshardware.interface.serial import Serial
from uoshardware.interface.stub import Stub


class Interface(Enum):
    """Enumerate interface module names."""

    STUB = Stub
    SERIAL = Serial



"""Module defining supported hardware definitions."""
from uoshardware.abstractions import UOS_SCHEMA, Device, Pin
from uoshardware.interface import Interface


# This is an interface for client implementations dead code false positive.
def enumerate_system_devices(interface_filter: Interface = None) -> []:  # dead: disable
    """Iterate through all interfaces and locates available devices.

    :param interface_filter: Interface enum to limit the search to a single interface type.
    :return: A list of uosinterface objects.
    """
    system_devices = []
    for interface in Interface:  # enum object
        if not interface_filter or interface_filter == interface:
            system_devices.extend(interface.value.enumerate_devices())
        if interface_filter is not None:
            break
    return system_devices


def get_device_definition(identity: str) -> Device:
    """Look up the system config dictionary for the defined device mappings.

    :param identity: String containing the lookup key of the device in the dictionary.
    :return: Device Object or None if not found
    """
    if identity is not None and identity.lower() in DEVICES:
        device = DEVICES[identity.lower()]
        for function_enabled in device.functions_enabled:
            device.functions_enabled[function_enabled] = {
                vol: UOS_SCHEMA[function_enabled].address_lut[vol]
                for vol in device.functions_enabled[function_enabled]
            }
    else:
        device = None
    return device


ARDUINO_NANO_3 = Device(
    name="Arduino Nano 3",
    interfaces=[Interface.STUB, Interface.SERIAL],
    functions_enabled={
        "set_gpio_output": {0: True},
        "get_gpio_input": {0: True},
        "get_adc_input": {0: True},
        "reset_all_io": {0: True},
        "hard_reset": {0: True},
        "get_system_info": {0: True},
        "get_gpio_config": {0: True},
    },
    digital_pins={
        2: Pin(gpio_out=True, gpio_in=True, pull_up=True, pc_int=True, hw_int=True),
        3: Pin(
            gpio_out=True,
            gpio_in=True,
            pull_up=True,
            pwm_out=True,
            pc_int=True,
            hw_int=True,
        ),
        4: Pin(gpio_out=True, gpio_in=True, pull_up=True, pc_int=True),
        5: Pin(gpio_out=True, gpio_in=True, pull_up=True, pwm_out=True, pc_int=True),
        6: Pin(
            gpio_out=True,
            gpio_in=True,
            pull_up=True,
            pwm_out=True,
            pc_int=True,
            comp={"type": "low", "bus": 0},
        ),
        7: Pin(
            gpio_out=True,
            gpio_in=True,
            pull_up=True,
            pc_int=True,
            comp={"type": "high", "bus": 0},
        ),
        8: Pin(gpio_out=True, gpio_in=True, pull_up=True, pc_int=True),
        9: Pin(gpio_out=True, gpio_in=True, pull_up=True, pwm_out=True, pc_int=True),
        10: Pin(
            gpio_out=True,
            gpio_in=True,
            pull_up=True,
            pwm_out=True,
            pc_int=True,
            spi={"type": "ss", "bus": 0},
        ),
        11: Pin(
            gpio_out=True,
            gpio_in=True,
            pull_up=True,
            pwm_out=True,
            pc_int=True,
            spi={"type": "mosi", "bus": 0},
        ),
        12: Pin(
            gpio_out=True,
            gpio_in=True,
            pull_up=True,
            pc_int=True,
            spi={"type": "miso", "bus": 0},
        ),
        13: Pin(
            gpio_out=True,
            gpio_in=True,
            pull_up=True,
            pc_int=True,
            i2c={"type": "sck", "bus": 0},
        ),
        14: Pin(
            gpio_out=True, gpio_in=True, pull_up=True, pc_int=True
        ),  # analogue pin 0
        15: Pin(
            gpio_out=True, gpio_in=True, pull_up=True, pc_int=True
        ),  # analogue pin 1
        16: Pin(
            gpio_out=True, gpio_in=True, pull_up=True, pc_int=True
        ),  # analogue pin 2
        17: Pin(
            gpio_out=True, gpio_in=True, pull_up=True, pc_int=True
        ),  # analogue pin 3
        18: Pin(
            gpio_out=True,
            gpio_in=True,
            pull_up=True,
            pc_int=True,
            i2c={"type": "sda", "bus": 0},
        ),  # analogue pin 4
        19: Pin(
            gpio_out=True,
            gpio_in=True,
            pull_up=True,
            pc_int=True,
            i2c={"type": "scl", "bus": 0},
        ),  # analogue pin 5
    },
    analogue_pins={
        0: Pin(adc_in=True),
        1: Pin(adc_in=True),
        2: Pin(adc_in=True),
        3: Pin(adc_in=True),
        4: Pin(adc_in=True),
        5: Pin(adc_in=True),
        6: Pin(adc_in=True),
        7: Pin(adc_in=True),
        8: Pin(adc_in=True),
    },
    aux_params={"default_baudrate": 115200},
)


DEVICES = {
    "hwid0": ARDUINO_NANO_3,
    "arduino_nano": ARDUINO_NANO_3,
}  # define aliases
